@@ -19,8 +19,8 @@ import gdb
 import os
 import os.path
 
-pythondir = '/gnu/store/eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee-gcc-static-5.4.0/aarch64-linux-gnu/share/gcc-5.4.0/python'
-libdir = '/gnu/store/eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee-gcc-static-5.4.0/lib'
+pythondir = '/gnu/store/eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee-gcc-static-5.5.0/powerpc-linux-gnu/share/gcc-5.5.0/python'
+libdir = '/gnu/store/eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee-gcc-static-5.5.0/lib'
 
 # This file might be loaded when there is no current objfile.  This
 # can happen if the user loads it manually.  In this case we don't
